@@ -2,9 +2,10 @@
 from __future__ import unicode_literals, print_function
 
 import uuid
+import hashlib
 
 from .contract import KongAdminContract, APIPluginConfigurationAdminContract, APIAdminContract, ConsumerAdminContract, \
-    PluginAdminContract
+    PluginAdminContract, BasicAuthAdminContract, KeyAuthAdminContract, OAuth2AdminContract
 from .utils import timestamp, uuid_or_string, add_url_params, filter_api_struct, filter_dict_list, assert_dict_keys_in, \
     ensure_trailing_slash
 from .compat import OrderedDict
@@ -17,6 +18,11 @@ class SimulatorDataStore(object):
         self._data_struct_filter = data_struct_filter or {}
         self._data = OrderedDict()
 
+    def destroy(self):
+        self.api_url = None
+        self._data_struct_filter = None
+        self._data = None
+
     def count(self):
         return len(self._data.keys())
 
@@ -48,10 +54,11 @@ class SimulatorDataStore(object):
             self._data[value_or_id].update(data_struct_update)
             return filter_api_struct(self._data[value_or_id], self._data_struct_filter)
 
-        for id in self._data:
-            if self._data[id][key] == value_or_id:
-                self._data[id].update(data_struct_update)
-                return filter_api_struct(self._data[id], self._data_struct_filter)
+        if key is not None:
+            for id in self._data:
+                if self._data[id][key] == value_or_id:
+                    self._data[id].update(data_struct_update)
+                    return filter_api_struct(self._data[id], self._data_struct_filter)
 
     def retrieve(self, value_or_id, key):
         value_or_id = uuid_or_string(value_or_id)
@@ -59,9 +66,10 @@ class SimulatorDataStore(object):
         if value_or_id in self._data:
             return filter_api_struct(self._data[value_or_id], self._data_struct_filter)
 
-        for id in self._data:
-            if self._data[id][key] == value_or_id:
-                return filter_api_struct(self._data[id], self._data_struct_filter)
+        if key is not None:
+            for id in self._data:
+                if self._data[id][key] == value_or_id:
+                    return filter_api_struct(self._data[id], self._data_struct_filter)
 
     def list(self, size, offset, **filter_fields):
         data_list = [filter_api_struct(data_struct, self._data_struct_filter)
@@ -99,10 +107,11 @@ class SimulatorDataStore(object):
         if value_or_id in self._data:
             del self._data[value_or_id]
 
-        for id in self._data:
-            if self._data[id][key] == value_or_id:
-                del self._data[id]
-                break
+        if key is not None:
+            for id in self._data:
+                if self._data[id][key] == value_or_id:
+                    del self._data[id]
+                    break
 
     def _get_by_field(self, field, value):
         for data_struct in self._data.values():
@@ -117,6 +126,12 @@ class APIPluginConfigurationAdminSimulator(APIPluginConfigurationAdminContract):
         self.api_url = api_url
         self._data = OrderedDict()
 
+    def destroy(self):
+        self.api_admin = None
+        self.api_name_or_id = None
+        self.api_url = None
+        self._data = None
+
     def create(self, plugin_name, enabled=None, consumer_id=None, **fields):
         plugins = PluginAdminSimulator.PLUGINS
 
@@ -240,6 +255,16 @@ class APIPluginConfigurationAdminSimulator(APIPluginConfigurationAdminContract):
                 del self._data[plugin_name]
                 break
 
+    def retrieve(self, plugin_name_or_id):
+        plugin_name_or_id = uuid_or_string(plugin_name_or_id)
+
+        if plugin_name_or_id in self._data:
+            return self._data[plugin_name_or_id]
+
+        for plugin_name in self._data:
+            if self._data[plugin_name]['id'] == plugin_name_or_id:
+                return self._data[plugin_name]
+
     def count(self):
         return len(self._data.keys())
 
@@ -255,10 +280,19 @@ class APIAdminSimulator(APIAdminContract):
             })
         self._plugin_admins = {}
 
+    def destroy(self):
+        self._store.destroy()
+        self._store = None
+
+        for key in self._plugin_admins:
+            self._plugin_admins[key].destroy()
+            del self._plugin_admins[key]
+        self._plugin_admins = None
+
     def count(self):
         return self._store.count()
 
-    def add(self, target_url, name=None, public_dns=None, path=None, strip_path=False):
+    def add(self, target_url, name=None, public_dns=None, path=None, strip_path=False, preserve_host=False):
         assert target_url is not None
         if not public_dns and not path:
             raise ValueError('At least a \'public_dns\' or a \'path\' must be specified, '
@@ -273,16 +307,19 @@ class APIAdminSimulator(APIAdminContract):
             'path': path,
             'target_url': target_url,
             'strip_path': strip_path,
+            'preserve_host': preserve_host,
             'created_at': timestamp()
         }, check_conflict_keys=('name', 'public_dns'))
 
-    def add_or_update(self, target_url, api_id=None, name=None, public_dns=None, path=None, strip_path=False):
+    def add_or_update(self, target_url, api_id=None, name=None, public_dns=None, path=None, strip_path=False,
+                      preserve_host=False):
         data = {
             'name': name or public_dns,
             'public_dns': public_dns,
             'path': path,
             'target_url': target_url,
-            'strip_path': strip_path
+            'strip_path': strip_path,
+            'preserve_host': preserve_host
         }
 
         if api_id is not None:
@@ -329,6 +366,158 @@ class APIAdminSimulator(APIAdminContract):
         return self._plugin_admins[api_id]
 
 
+class BasicAuthAdminSimulator(BasicAuthAdminContract):
+    def __init__(self, consumer_admin, consumer_id, api_url):
+        self.consumer_admin = consumer_admin
+        self.consumer_id = consumer_id
+        self._store = SimulatorDataStore(api_url or 'http://localhost:8001/consumers/%s/basicauth' % self.consumer_id)
+
+    def destroy(self):
+        self.consumer_admin = None
+        self.consumer_id = None
+
+        self._store.destroy()
+        self._store = None
+
+    def create_or_update(self, basic_auth_id=None, username=None, password=None):
+        data = {
+            'username': username,
+            'password': password
+        }
+
+        if basic_auth_id is not None:
+            return self.update(basic_auth_id, **data)
+
+        return self.create(**data)
+
+    def create(self, username, password):
+        assert username and password
+
+        return self._store.create({
+            'username': username,
+            'password': password,
+            'created_at': timestamp()
+        }, check_conflict_keys=('username',))
+
+    def update(self, basic_auth_id, **fields):
+        return self._store.update(basic_auth_id, None, fields)
+
+    def list(self, size=100, offset=None, **filter_fields):
+        return self._store.list(size=size, offset=offset, **filter_fields)
+
+    def delete(self, basic_auth_id):
+        return self._store.delete(basic_auth_id, None)
+
+    def retrieve(self, basic_auth_id):
+        return self._store.retrieve(basic_auth_id, None)
+
+    def count(self):
+        return self._store.count()
+
+
+class KeyAuthAdminSimulator(KeyAuthAdminContract):
+    def __init__(self, consumer_admin, consumer_id, api_url):
+        self.consumer_admin = consumer_admin
+        self.consumer_id = consumer_id
+        self._store = SimulatorDataStore(api_url or 'http://localhost:8001/consumers/%s/keyauth' % self.consumer_id)
+
+    def destroy(self):
+        self.consumer_admin = None
+        self.consumer_id = None
+
+        self._store.destroy()
+        self._store = None
+
+    def create_or_update(self, key_auth_id=None, key=None):
+        data = {
+            'key': key or self._generate_key()
+        }
+
+        if key_auth_id is not None:
+            return self.update(key_auth_id, **data)
+
+        return self.create(**data)
+
+    def create(self, key=None):
+        return self._store.create({
+            'key': key or self._generate_key(),
+            'created_at': timestamp()
+        }, check_conflict_keys=('key',))
+
+    def update(self, key_auth_id, **fields):
+        return self._store.update(key_auth_id, None, fields)
+
+    def list(self, size=100, offset=None, **filter_fields):
+        return self._store.list(size=size, offset=offset, **filter_fields)
+
+    def delete(self, key_auth_id):
+        return self._store.delete(key_auth_id, None)
+
+    def retrieve(self, key_auth_id):
+        return self._store.retrieve(key_auth_id, None)
+
+    def count(self):
+        return self._store.count()
+
+    def _generate_key(self):
+        data = str(uuid.uuid4()).encode('utf-8')
+
+        m = hashlib.sha1()
+        m.update(data)
+        return m.hexdigest()
+
+
+class OAuth2AdminSimulator(OAuth2AdminContract):
+    def __init__(self, consumer_admin, consumer_id, api_url):
+        self.consumer_admin = consumer_admin
+        self.consumer_id = consumer_id
+        self._store = SimulatorDataStore(api_url or 'http://localhost:8001/consumers/%s/oauth2' % self.consumer_id)
+
+    def destroy(self):
+        self.consumer_admin = None
+        self.consumer_id = None
+
+        self._store.destroy()
+        self._store = None
+
+    def create_or_update(self, oauth2_id=None, name=None, redirect_uri=None, client_id=None, client_secret=None):
+        data = {
+            'name': name,
+            'redirect_uri': redirect_uri,
+            'client_id': client_id,
+            'client_secret': client_secret
+        }
+
+        if oauth2_id is not None:
+            return self.update(oauth2_id, **data)
+
+        return self.create(**data)
+
+    def create(self, name, redirect_uri, client_id=None, client_secret=None):
+        assert name and redirect_uri
+
+        return self._store.create({
+            'name': name,
+            'redirect_uri': redirect_uri,
+            'created_at': timestamp()
+        }, check_conflict_keys=('name', 'redirect_uri'))
+
+    def update(self, oauth2_id, **fields):
+        return self._store.update(oauth2_id, None, fields)
+
+    def list(self, size=100, offset=None, **filter_fields):
+        return self._store.list(size=size, offset=offset, **filter_fields)
+
+    def delete(self, oauth2_id):
+        return self._store.delete(oauth2_id, None)
+
+    def retrieve(self, oauth2_id):
+        return self._store.retrieve(oauth2_id, None)
+
+    def count(self):
+        return self._store.count()
+
+
 class ConsumerAdminSimulator(ConsumerAdminContract):
     def __init__(self, api_url=None):
         self._store = SimulatorDataStore(
@@ -337,6 +526,22 @@ class ConsumerAdminSimulator(ConsumerAdminContract):
                 'custom_id': None,
                 'username': None
             })
+        self._basic_auth_admins = {}
+        self._key_auth_admins = {}
+        self._oauth2_admins = {}
+
+    def destroy(self):
+        self._store.destroy()
+        self._store = None
+
+        for related_admin in (self._basic_auth_admins, self._key_auth_admins, self._oauth2_admins):
+            for key in related_admin:
+                related_admin[key].destroy()
+                del related_admin[key]
+
+        self._basic_auth_admins = None
+        self._key_auth_admins = None
+        self._oauth2_admins = None
 
     def count(self):
         return self._store.count()
@@ -371,8 +576,54 @@ class ConsumerAdminSimulator(ConsumerAdminContract):
         return self._store.list(size, offset, **filter_fields)
 
     def delete(self, username_or_id):
+        consumer_id = self.retrieve(username_or_id).get('id')
+
+        if consumer_id is None:
+            raise ValueError('Unknown username_or_id: %s' % username_or_id)
+
+        if consumer_id in self._basic_auth_admins:
+            self._basic_auth_admins[consumer_id].consumer_admin = None
+            del self._basic_auth_admins[consumer_id]
+
+        if consumer_id in self._oauth2_admins:
+            self._oauth2_admins[consumer_id].consumer_admin = None
+            del self._oauth2_admins[consumer_id]
+
         return self._store.delete(username_or_id, 'username')
 
+    def basic_auth(self, username_or_id):
+        consumer_id = self.retrieve(username_or_id).get('id')
+
+        if consumer_id is None:
+            raise ValueError('Unknown username_or_id: %s' % username_or_id)
+
+        if consumer_id not in self._basic_auth_admins:
+            self._basic_auth_admins[consumer_id] = BasicAuthAdminSimulator(self, consumer_id, self._store.api_url)
+
+        return self._basic_auth_admins[consumer_id]
+
+    def key_auth(self, username_or_id):
+        consumer_id = self.retrieve(username_or_id).get('id')
+
+        if consumer_id is None:
+            raise ValueError('Unknown username_or_id: %s' % username_or_id)
+
+        if consumer_id not in self._key_auth_admins:
+            self._key_auth_admins[consumer_id] = KeyAuthAdminSimulator(self, consumer_id, self._store.api_url)
+
+        return self._key_auth_admins[consumer_id]
+
+    def oauth2(self, username_or_id):
+        consumer_id = self.retrieve(username_or_id).get('id')
+
+        if consumer_id is None:
+            raise ValueError('Unknown username_or_id: %s' % username_or_id)
+
+        if consumer_id not in self._oauth2_admins:
+            self._oauth2_admins[consumer_id] = OAuth2AdminSimulator(self, consumer_id, self._store.api_url)
+
+        return self._oauth2_admins[consumer_id]
+
 
 class PluginAdminSimulator(PluginAdminContract):
     # Copied from real kong server, v0.4.0
@@ -427,6 +678,9 @@ class PluginAdminSimulator(PluginAdminContract):
         'requestsizelimiting': {'fields': {'allowed_payload_size': {'default': 128, 'type': 'number'}}}
     })
 
+    def destroy(self):
+        pass
+
     def list(self):
         return {
             'enabled_plugins': self.PLUGINS.keys()
@@ -442,3 +696,8 @@ class KongAdminSimulator(KongAdminContract):
             apis=APIAdminSimulator(api_url=api_url),
             consumers=ConsumerAdminSimulator(api_url=api_url),
             plugins=PluginAdminSimulator())
+
+    def close(self):
+        self.apis.destroy()
+        self.consumers.destroy()
+        self.plugins.destroy()
